/*
 *   ALSA sequencer FIFO
 *   Copyright (c) 1998 by Frank van de Pol <frank@vande-pol.demon.nl>
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
#ifndef __SND_SEQ_FIFO_H
#define __SND_SEQ_FIFO_H

#include "seq_memory.h"


/* === FIFO === */

typedef struct {
	pool_t *pool;			/* FIFO pool */
	snd_seq_event_cell_t* head;    	/* pointer to head of fifo */
	snd_seq_event_cell_t* tail;    	/* pointer to tail of fifo */
	int cells;
	spinlock_t lock;
	wait_queue_head_t input_sleep;
	struct semaphore use_mutex;
	int overflow;

} fifo_t;

/* create new fifo (constructor) */
extern fifo_t *snd_seq_fifo_new(int poolsize);

/* delete fifo (destructor) */
extern void snd_seq_fifo_delete(fifo_t **f);


/* enqueue event to fifo */
extern int snd_seq_fifo_event_in(fifo_t *f, snd_seq_event_t *event, int atomic);

/* dequeue cell from fifo */ 
extern snd_seq_event_cell_t *snd_seq_fifo_cell_out(fifo_t *f);
extern snd_seq_event_cell_t *snd_seq_fifo_locked_cell_out(fifo_t *f, int *err, int nonsleep);

/* free dequeued cell */
extern void snd_seq_fifo_cell_putback(fifo_t *f, snd_seq_event_cell_t *cell);

/* unlock FIFO */
extern void snd_seq_fifo_unlock(fifo_t *f);

/* clean up queue */
extern void snd_seq_fifo_clear(fifo_t *f);

/* polling */
extern int snd_seq_fifo_poll_wait(fifo_t *f, struct file *file, poll_table *wait);

/* resize pool in fifo */
int snd_seq_fifo_resize(fifo_t *f, int poolsize);


#endif
