/*
 *  CLModem v0.3.0 - Low level DSP routines
 *
 *  Copyright (C) 2000  Mikhail Moreyra
 *
 *  This file may be redistributed under the terms of the GNU Public
 *  License.
 */

#define __NO_VERSION__

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/wrapper.h>
#include <asm/io.h>
#include <asm/irq.h>

#include "clm_config.h"
#include "clm_data.h"
#include "clm_dspdrv.h"
#include "clm_modem.h"


#define BUF_SIZE	4096

extern int BaseIOAddress;
extern int modem_irq;

unsigned char current_command[40];
unsigned char dsp_response_data[16];
unsigned char Cdata[64];

int hook = 1;
int spkr_volume_level = 0;
int spkr_on = 0;

int _dsp_timer;
volatile int dsp_done;
int mt_x2_connection;
int mt_v90_connection;
int ack_or_response;
int mt_v90_configured;
volatile int v8_configured = 0;

unsigned char line_mode = 0;
unsigned short mt_RateSeq1;
unsigned short mt_RateSeq2;
unsigned short mt_RateSeq3;
unsigned short mt_rate_seq_map, mt_set_rate_seq_param3;

int mt_conn_orig = 1;
int line_bit_rate;
int call_progress_state = 0;
int PcmEnabled = 0;
int mt_IdleSendPwron = 0;
volatile int mt_RingState = 0;
char mt_call_progress;

volatile char line_io_enabled = 0;
volatile char dce_length_semaphore = 0;
volatile char dce_status_semaphore = 0;
volatile int dce_tx_count = 0;
volatile int dce_rx_count = 0;
volatile int dce_tx_interrupt_expected = 0;
volatile unsigned char dce_tx_buffer[BUF_SIZE];
volatile unsigned char dce_rx_buffer[BUF_SIZE];
volatile int dce_tx_remove = 0;
volatile int dce_rx_insert = 0;
volatile int dte_rx_remove = 0;
volatile int dte_tx_insert = 0;
volatile int stat_tx_count = 0;
volatile int stat_rx_count = 0;
int dce_rx_errors = 0;

unsigned char v8_tx_buffer[32];
int sending_CM, sending_JM;
int JM_received, CM_received;
int CM_count, JM_count;
unsigned char v8_tx_msg_size;
volatile int jm_seq_match;
int v8_status, length, old_length;
int ci_count, v8_rx_msg_size;

int RX_stat[32];
int TX_stat[32];

volatile int critical = 0;

/* ************************************************ */
void cmd_packet_callback(unsigned char x);
void mt_CallProgressCallback(void);
void line_int(void);
void line_rx_int(void);
void line_tx_int(void);
void sync_rx_int(void);
void sync_tx_int(void);
void v8_rx_int(void);
void v8_tx_int(void);
/* ************************************************ */

void
set_register(int reg, unsigned char val)
{
    outb (val, BaseIOAddress + reg);
}

unsigned char
get_register(int reg)
{
    return inb (BaseIOAddress + reg);
}

/*************************************************/
/*		INTERRUPT HANDLER		 */
/*************************************************/

void
dspdrv_CommRamHandler(void)
{
    unsigned char x, rv;

    rv = get_register (0x0);

    if (rv) {
	x = dspdrv_GetCRAM (Cdata);
	cmd_packet_callback (x);
    }
}

void
dspdrv_CommRamISR(void)
{
    critical = 1;
    get_register (0xff);
    dspdrv_CommRamHandler();
    line_int();
    critical = 0;
}

void
cmd_packet_callback (unsigned char x)
{
    if (x == 0xd1)
	mt_CallProgressCallback();
    else
	dsp_done = 1;
}

void
mt_CallProgressCallback(void)
{
    static short LastTimeStamp = 0;
    short nt, xt, dt;
    int x;

    if (Cdata[4] != 1) return;
    nt = (Cdata[7] << 8) | Cdata[6];
    xt = (Cdata[9] << 8) | Cdata[8];
    if (nt > LastTimeStamp)
	dt = nt - LastTimeStamp;
    else {
	dt = nt - LastTimeStamp;
	dt += 0x1aab;
    }
    /* Dbg("Ring time: %2d - dtime: %4d -- ", xt, dt); */
    if (xt) {
	if (dt)
	    x = (xt * 0x3e8) / dt;
	else
	    x = 0;
	if ((x >= 0xf) && (x <= 0x44))
	    mt_RingState = 7;
	else
	    mt_RingState = 0;
	/* Dbg("Ring length= %d -- ", x); */
    } else {
	if (mt_RingState == 2)
	    mt_RingState = 0;
	else
	    mt_RingState = 2;
    }
    Msg("RingState = %d\n", mt_RingState);
    LastTimeStamp = nt;
}

/* CRAM_CallBack */
void
line_int(void)
{
    if (line_io_enabled == 0) return;

    dspdrv_ReadCram(0x34, &dce_length_semaphore, 1);
    dspdrv_ReadCram(0x35, &dce_status_semaphore, 1);
    if (dce_status_semaphore)
	dspdrv_ReadCram(0x34, &dce_length_semaphore, 1);
#if 1
    if (dce_status_semaphore>1) Dbg("RX StSem: %#x\n", dce_status_semaphore);
#endif
    if (dce_length_semaphore || dce_status_semaphore)
	line_rx_int();
    dspdrv_ReadCram(0x74, &dce_length_semaphore, 1);
    dspdrv_ReadCram(0x75, &dce_status_semaphore, 1);
#if 1
    if (dce_status_semaphore>1) Dbg("TX StSem: %#x\n", dce_status_semaphore);
#endif
    if ((dce_length_semaphore == 0) && (dce_status_semaphore == 0))
	line_tx_int();
}


void
line_rx_int(void)
{
    if (v8_configured) v8_rx_int();
    else sync_rx_int();
}

void
line_tx_int(void)
{
    if (v8_configured) v8_tx_int();
    else sync_tx_int();
}

void
sync_rx_int(void)
{
    char reg = 0x14;

    if (dce_status_semaphore == 6) {
	/* Msg("RX Error\n"); */
	dce_rx_errors++;
    }
    
    RX_stat[(int)dce_length_semaphore-1]++;
    if (dce_length_semaphore > 32) Dbg("ERROR: dce_length_semaphore > 32\n");

    while (dce_length_semaphore) {
	dspdrv_ReadCram(reg++, &dce_rx_buffer[dce_rx_insert], 1);
	if (dce_rx_count < BUF_SIZE) {
	    dce_rx_insert++;
	    dce_rx_count++;
	    stat_rx_count++;
	    if (dce_rx_insert == BUF_SIZE) dce_rx_insert = 0;
	} else Dbg("ERROR: RX buffer full\n");
	dce_length_semaphore--;
    }
    dce_status_semaphore = 0;
    dspdrv_WriteCram(0x34, &dce_length_semaphore, 1);
    dspdrv_WriteCram(0x35, &dce_status_semaphore, 1);
}

void
sync_tx_int(void)
{
    char reg, xs, tcnt;

    if (dce_tx_count == 0) {
	dce_tx_interrupt_expected = 0;
	return;
    }
    reg = 0x54;
    tcnt = 0;
    while (tcnt < 0x20) {	/* 32 Tx bytes Max */
	dspdrv_WriteCram(reg++, &dce_tx_buffer[dce_tx_remove++], 1);
	dce_tx_count--;
	tcnt++;
	stat_tx_count++;
	if (dce_tx_remove == BUF_SIZE) dce_tx_remove = 0;
	if (dce_tx_count == 0) break;
    }
    TX_stat[(int)tcnt-1]++;
    
    xs = 0;
    dspdrv_WriteCram(0x75, &xs, 1);
    dspdrv_WriteCram(0x74, &tcnt, 1);
}

void
v8_rx_int(void)
{
    unsigned char c, reg = 0x14;

    while (dce_length_semaphore) {
	dspdrv_ReadCram(reg++, &c, 1);
	dce_length_semaphore--;
	/* Dbg("%4x ", c); */
	if (v8_status == 0) {
	    if (c == 0xe0) {
		v8_status = 1;
		old_length = 0;
		dce_rx_buffer[0] = 0xe0;
		length = 1;
	    }
	} else if (v8_status == 1) {
	    if (c != 0xe0)
		dce_rx_buffer[length++] = c;
	    else if (length < 3)
		length = 1;
	    else {
		v8_status = 2;
		old_length = length - 1;
		length = 1;
	    }
	} else if (v8_status == 2) {
	    if (dce_rx_buffer[length] == c) {
		length++;
		old_length--;
	    } else if (c != 0xe0) {
		v8_status = 1;
		dce_rx_buffer[length++] = c;
	    } else {
		if (old_length == 0) {
		    /* Dbg("JM Match\n"); */
		    v8_rx_msg_size = length;
		    jm_seq_match = 1;
		    dce_length_semaphore = 0;
		    v8_status = 3;
		} else {
		    Dbg("Old lenght not 0...\n");
		    if (length < 3) {
			length = 1;
			v8_status = 1;
		    } else {
			old_length = length - 1;
			length = 1;
		    }
		}
	    }
	} else {
	    dce_length_semaphore = 0;
	}
    }
    dspdrv_WriteCram(0x34, &dce_length_semaphore, 1);
    dce_status_semaphore = 0;
    dspdrv_WriteCram(0x35, &dce_status_semaphore, 1);
}

void
v8_tx_int(void)
{
    char xs = 0;

    if (!sending_CM && !sending_JM) return;
    dspdrv_WriteCram(0x54, v8_tx_buffer, v8_tx_msg_size);
    dspdrv_WriteCram(0x75, &xs, 1);
    dspdrv_WriteCram(0x74, &v8_tx_msg_size, 1);
    if (sending_CM) CM_count++;
    else JM_count++;
}

void
line_enable_interrupts(void)
{
    enable_irq(modem_irq);
}

void
line_disable_interrupts(void)
{
    disable_irq(modem_irq);
}

/* ************************************************ */
void
delay(int milliseconds)
{
    /* mdelay(milliseconds);     This doesn't work well */
    
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout((milliseconds * HZ + 1000 - HZ) / 1000); /* 100ths sec */
}
/* ************************************************ */

void
dspdrv_clear_dsp_interrupt(void)
{
    get_register (0xff);
}

void
dspdrv_ping_cram_intr(void)
{
    set_register (0xfe, 0);
}

void
SetDspCfg (int p1)
{
    unsigned char rv;
    
    rv = get_register (0xf6);
    if (p1 == 0) rv &= 0xfe;
    else rv |= 0x1;
    set_register (0xf6, rv);
}

void
WaitForDspReady(void)
{
    /* delay(1000); */
    delay(500);   /* How much should this be ? */
}

void
ResetDspInternal(void)
{
    set_register (0xfc, 2);
    set_register (0xf7, 0);
    SetDspCfg (1);
    set_register (0xfc, 1);
    WaitForDspReady();
}

#define LO_B(x) ((unsigned short)x & 0xff)
#define HI_B(x) (((unsigned short)x >> 8) & 0xff)

int
DnldViaHw(unsigned short *data)
{
    unsigned short xc, wc, w;
    unsigned char xl=0, xh=0;
    int i, rv=1;
    
    xc = data[0];
    i = 1;
    set_register(0xfc, 0xa);
    while (xc) {
	wc = data[i++];
	set_register(0xf9, HI_B(data[i]));
	set_register(0xf8, LO_B(data[i]));
	i++;
	while (wc) {
	    set_register(0xfa, LO_B(data[i]));
	    set_register(0xfa, HI_B(data[i]));
	    i++;
	    wc--;
	}
	xc--;
    }
    i = 1;
    xc = data[0];
    while (xc) {
	wc = data[i++];
	xl = LO_B(data[i]);
	xh = HI_B(data[i]);
	set_register(0xf8, xl);
	set_register(0xf9, xh);
	i++;
	rv = 1;
	while (wc) {
	    w = get_register(0xfa) | (get_register(0xfa) << 8);
	    if (data[i] != w) rv = 0;
	    i++;
	    wc--;
	}
	xc--;
    }
    set_register(0xf9, 0x8);
    set_register(0xf8, 0);
    set_register(0xfa, xl);
    set_register(0xfa, xh);
    set_register(0xfc, 0x9);
    return rv;
}


int
DnldBootx(void) {
    /* return DnldViaHw(bootx_array); */
    return 0;
}

int
DnldPcmPage(void)
{
    return 0;	/* No PCM modes supported yet. (x2 & V.90) */
}

void
DnldPoweronPatch(void)
{
    DnldViaHw(pwr_on_patch);
    WaitForDspReady();
    DspPatch(1);
}

void
ResetDspPoweron(void)
{
    int al;
    set_register (0xf7, 0x2);
    SetDspCfg (1);

    if (DnldBootx() && DnldPcmPage()) {
	if (mt_v90_configured == 0)
	    al = 0;
	else
	    al = 1;
	PcmEnabled = al+1;
    }
    DnldPoweronPatch();
}

void
DnldPatch(unsigned char *patch)
{
    int i, ps, pcnt, j, wx, b;
    
    ps = (patch[2] | (patch[3] << 8))*2 + 1;
    i = 2;
    ack_or_response = 2;
    current_command[0] = 0xc;
    current_command[1] = 0;
    current_command[3] = 0;
    current_command[6] = 0;
    current_command[7] = 0;
    while (i != ps) {
	pcnt = 4;
	j = 8;
	current_command[4] = patch[i*2];
	current_command[5] = patch[i*2+1];
	wx = patch[i*2] | (patch[i*2+1] << 8);
	for (b=1; b<4; b++) {
	    i++;
	    pcnt+=2;
	    current_command[j++] = patch[i*2];
	    current_command[j++] = patch[i*2+1];
	    if (i == ps) break;
	    i++;
	    wx++;
	    if (wx != (patch[i*2] | (patch[i*2+1] << 8))) break;
	}
	current_command[2] = pcnt;
	send_pkt_to_dsp();
    }
}

void
ConfigPatch (unsigned char p1)
{
    if (mt_x2_connection || mt_v90_connection) return;
    current_command[0] = 0x9a;
    current_command[1] = 0;
    current_command[2] = 2;
    current_command[3] = 0;
    current_command[4] = p1;
    current_command[5] = 0;
    ack_or_response = 0;
    send_pkt_to_dsp ();
}

void
DspPatch (int n)
{
    void *patch;
    
    if (mt_x2_connection || mt_v90_connection) return;
    if (n == 1) {
	DnldPatch (common_patch);
	//patch = v90_com_patch;
	patch = data_patch;	/* FIXME */
    }
    else if (n == 2) patch = data_patch;
    else if ((n == 3) || (n == 4)) patch = training_patch;
    else patch = voice_patch;		/* 5,6 */
    /* if (((n == 1) && mt_v90_configured) || ((n>1) && (n<=6))) */
    if ((n>1) && (n<=6))
	DnldPatch (patch);
    WriteToDsp (0x17e4, 0x7c00);
}

void
ResetDspIdle(void)
{
    line_bit_rate = 0;
    mt_x2_connection = 0;
    mt_v90_connection = 0;
    call_progress_state = 0;
    
    if (mt_IdleSendPwron) {
	DnldPoweronPatch ();
	set_hook_relay (0);
	set_voice_relay (0);
	set_cid_relay (0);
	set_spk_hs_relay (0);
    } else mt_IdleSendPwron = 1;	// FIXME TESTING
    current_command[0] = 0x60;
    current_command[1] = 0;
    current_command[2] = 0;
    current_command[3] = 0;
    ack_or_response = 0;
    send_pkt_to_dsp ();
    ConfigPatch (5);
    DspPatch (1);
    DspPatch (5);
    ConfigPatch (3);
}

/**************************************************************/
int
call_progress(void)
{
    unsigned char preg;
    
    preg = read_dsp_call_progress_reg ();
    if ((preg == 7) || (preg == 2)) preg = 0;
    if (mt_RingState) {
	preg = mt_RingState;
	if (mt_RingState == 7) mt_RingState = 0;
    }
    return preg;
}

unsigned char
ReadDspPcmStatusReg(void)
{
    unsigned char x;
    dspdrv_ReadCram (0x3c, &x, 1);
    return x;
}

unsigned char
read_dsp_modem_status_reg(void)
{
    unsigned char x;
    dspdrv_ReadCram (0x36, &x, 1);
    return x;
}

unsigned char
read_dsp_call_progress_reg(void)
{
    dspdrv_ReadCram (0x38, &mt_call_progress, 1);
    return mt_call_progress;
}

/******************************************************** */
void
set_hook_relay(int n)
{
    unsigned char rv;
    
    rv = get_register(0xf0) | 0x4;
    set_register (0xf0, rv);
    rv = get_register(0xf3) & 0xfb;
    set_register (0xf3, rv);
    rv = get_register(0xf2);
    
    if (n == 1) rv &= 0xfb;
    else rv |= 0x4;

    set_register (0xf2, rv);
}

void
set_voice_relay(int n)
{
    unsigned char rv;
    
    rv = get_register(0xf0) | 0x40;
    set_register (0xf0, rv);
    rv = get_register(0xf3) & 0xbf;
    set_register (0xf3, rv);
    rv = get_register(0xf2);
    
    if (n == 1) rv &= 0xbf;
    else rv |= 0x40;

    set_register (0xf2, rv);
}

void
set_cid_relay(int n)
{
    unsigned char rv;
    
    rv = get_register(0xf0) | 0x80;
    set_register (0xf0, rv);
    rv = get_register(0xf3) & 0x7f;
    set_register (0xf3, rv);
    rv = get_register(0xf2);
    
    if (n == 1) rv &= 0x7f;
    else rv |= 0x80;

    set_register (0xf2, rv);
}

void
set_dsvd_relay(int n)
{
    unsigned char rv;
    
    rv = get_register(0xf0) | 0x8;
    set_register (0xf0, rv);
    rv = get_register(0xf3) & 0xf7;
    set_register (0xf3, rv);
    rv = get_register(0xf2);
    
    if (n == 1) rv &= 0xf7;
    else rv |= 0x8;

    set_register (0xf2, rv);
}

void
set_spk_hs_relay(int n)
{
    unsigned char rv;
    
    rv = get_register(0xf0) | 0x10;
    set_register (0xf0, rv);
    rv = get_register(0xf3) & 0xef;
    set_register (0xf3, rv);
    rv = get_register(0xf2);
    
    if (n == 1) rv &= 0xef;
    else rv |= 0x10;

    set_register (0xf2, rv);
}

int get_hook_relay(void)
{
    return (~get_register(0xf2) >> 2) & 0x1;
}

int get_voice_relay(void)
{
    return (~get_register(0xf2) >> 6) & 0x1;
}

int get_cid_relay (void)
{
    return (~get_register(0xf2) >> 7) & 0x1;
}

int get_dsvd_relay (void)
{
    return (~get_register(0xf2) >> 3) & 0x1;
}

int get_phone_relay (void)
{
    return  (~ReadFromDsp (0x69) >> 1) & 0x1;
}

int get_spk_hs_relay (void)
{
    return (~get_register(0xf2) >> 4) & 0x1;
}

/* ********************************************** */
void
dspdrv_PutCRAM (unsigned char *data)
{
    unsigned char l1, l2 ,i;
    l1 = data[2] + 4;
    if (l1 & 0x1) l1++;
    l2 = 1;
    if (l1 > 1) {
	i = 1;
	do {
	    set_register (l2 + 0x40, data[i++]);
	    l2++;
	} while (l2 < l1);
    }
    set_register (0x40, data[0]);
    set_register (0xfe, 1);
}

unsigned char
dspdrv_GetCRAM (unsigned char *data)
{
    unsigned char l1, l2 ,ret;
    int x;
    data[0] = ret = get_register (0x0);
    data[1] = get_register (0x1);
    data[2] = l1 = get_register (0x2);
    data[3] = get_register (0x3);
    l2 = 4;

    if (l1 > 0x3c) l1 = 0x3c;
    if (l1) {
	x = 4;
	do {
	    data[x++] = get_register (l2++);
	    l1--;
	} while (l1 > 0);
    }
    set_register (0x0, 0);
    set_register (0xfe, 1);
    return ret;
}

void
dspdrv_WriteCram (int ireg, volatile unsigned char *data, int count)
{
    while (count) {
	set_register (ireg++, *data);
	data++;
	count--;
    }
}

void
dspdrv_ReadCram (int ireg, volatile unsigned char *data, int count)
{
    while (count) {
	*data = get_register (ireg++);
	data++;
	count--;
    }
}

void
WriteToDsp (unsigned short p1, unsigned short p2)
{
    current_command[1] = 0;
    current_command[3] = 0;
    current_command[4] = p1 & 0xff;
    current_command[6] = 0;
    current_command[7] = 0;
    ack_or_response = 2;
    current_command[0] = 0xc;
    current_command[2] = 0x6;
    current_command[5] = (p1>>8) & 0xff;
    current_command[8] = p2 & 0xff;
    current_command[9] = (p2>>8) & 0xff;
    send_pkt_to_dsp ();
}

unsigned short
ReadFromDsp (unsigned short p1)
{
    current_command[0] = 0x9;
    current_command[1] = 0;
    current_command[2] = 0x4;
    current_command[3] = 0;
    current_command[4] = p1 & 0xff;
    current_command[5] = (p1>>8) & 0xff;
    current_command[6] = 0;
    current_command[7] = 0;
    current_command[8] = 0;
    ack_or_response = 1;
    send_pkt_to_dsp ();
    return (dsp_response_data[1]<<8) | dsp_response_data[0];
}

int
cp_dsp(void)
{
    int i = 0x32;

    dsp_done = 0;
    while (1) {
	_dsp_timer = 2;
	while (_dsp_timer) {
	    if (get_register (0x40)) continue;
	    dspdrv_PutCRAM (current_command);
	    return 1;
	}
	i--;
	if (i == 0) {
	    /* fatal_error(0x4e464, 0x4fb); */
	    return -1;
	}
    }
}

int
dsp_cp(void)
{
    int cnt;
    int i = 0x32;

    while (1) {
	_dsp_timer = 2;
	while (_dsp_timer) {
	    if (dsp_done == 1) {
		int x;
		cnt = Cdata[2];
		for (x=0; x < cnt; x++)
		    dsp_response_data[x] = Cdata[4+x];
		return 1;
	    }
	}
	i--;
	if (i == 0) {
	    /* fatal_error(0x4e4dc, 0x522); */
	    return -1;
	}
    }
}

void
send_pkt_to_dsp(void)
{
    int i;
    
    _dsp_timer = 0x64;
    if (current_command[2] % 2) current_command[2]++;
    while (1) {
	if (_dsp_timer) {
	    /* process_timers (); */
	    if (cp_dsp() != 1) continue;
	    if (_dsp_timer) break;
	}
	break;
    }
    if (ack_or_response != 2) {
	_dsp_timer = 0x64;
	do {
	    if (_dsp_timer == 0) break;
	    /* process_timers (); */
	} while ( dsp_cp() != 1);
    }

    /* Clear command buffer */
    for (i=4; i < 40; i++) current_command[i] = 0;
    /* update_modem_status(); */
}

/*******************************************************/
void
on_hook(void)
{
    if (hook == 0) {
	hook = 1;
	set_hook_relay(0);
	set_speaker(0);
    }
}

void
off_hook(void)
{
    if (hook) {
	hook = 0;
	set_hook_relay(1);
    }
}

/********************************************************/
void
dsp_info () {
    unsigned char al = get_register (0xf6);

    if (al & 0x1)
	Msg("DSP Code Location : Internal ROM\n");
    else
	Msg("DSP Code Location : External SRAM\n");
}

/************************************************************/
void
set_speaker(int n)
{
    char svol, x;
    
    if (n == 1) {
	svol = spkr_volume_level;
	spkr_on = 1;
    } else {
	spkr_on = 0;
	svol = 0;
    }
    current_command[0] = 0x81;
    current_command[1] = 0;
    current_command[2] = 2;
    current_command[3] = 0;
    current_command[4] = svol;
    current_command[5] = 0;
    ack_or_response = 0;
    send_pkt_to_dsp ();

    current_command[0] = 0xa;
    current_command[1] = 0;
    current_command[2] = 2;
    current_command[3] = 0;
    current_command[4] = 0x33;
    current_command[5] = 0;
    ack_or_response = 1;
    send_pkt_to_dsp ();

    x = dsp_response_data[1] & 0xc3;

    if (spkr_on) {
	if (spkr_volume_level <= 1) x |= 0x34;
	else if ((spkr_volume_level==2) || (spkr_volume_level != 3)) x|=0x38;
	else x |= 0x3c;
    } else x |= 0x30;
    
    current_command[0] = 0xd;
    current_command[1] = 0;
    current_command[2] = 4;
    current_command[3] = 0;
    current_command[4] = 0x33;
    current_command[5] = 0;
    current_command[6] = 0;
    current_command[7] = x;
    current_command[0xb] = 0;
    ack_or_response = 2;
    send_pkt_to_dsp ();
}

void
set_volume(int n)
{
    char x;
    
    if (n == 0) n = 1;
    spkr_volume_level = n;
    if (!spkr_on) return;

    current_command[0] = 0x81;
    current_command[1] = 0;
    current_command[2] = 2;
    current_command[3] = 0;
    current_command[4] = n;
    current_command[5] = 0;
    ack_or_response = 0;
    send_pkt_to_dsp ();

    current_command[0] = 0xa;
    current_command[1] = 0;
    current_command[2] = 2;
    current_command[3] = 0;
    current_command[4] = 0x33;
    current_command[5] = 0;
    ack_or_response = 1;
    send_pkt_to_dsp ();

    x = dsp_response_data[1] & 0xc3;

    if (spkr_volume_level <= 1) x |= 0x34;
    else if ((spkr_volume_level==2) || (spkr_volume_level != 3)) x|=0x38;
    else x |= 0x3c;
        
    current_command[0] = 0xd;
    current_command[1] = 0;
    current_command[2] = 4;
    current_command[3] = 0;
    current_command[4] = 0x33;
    current_command[5] = 0;
    current_command[6] = 0;
    current_command[7] = x;
    current_command[0xb] = 0;
    ack_or_response = 2;
    send_pkt_to_dsp ();
}

void
modem_config_dial(char t1, char t2, char pd, short p4)
{
    current_command[0] = 0x66;
    current_command[1] = 0;
    current_command[2] = 6;
    current_command[3] = 0;
    if (pd == 1) {	/* PULSE DIALING */
	current_command[4] = t2;
	current_command[5] = t1;
    } else {		/* TONE DIALING */
	current_command[4] = t1;
	current_command[5] = t2;
    }
    current_command[6] = pd;
    current_command[7] = 0;
    current_command[8] = p4 & 0xff;
    current_command[9] = (p4 >> 8) & 0xff;
   
    ack_or_response = 0;
    send_pkt_to_dsp ();
}

void
modem_dial_digit(char digit)
{
    char x;
 
    if (digit == '#') x = 0xf;
    else if (digit == '*') x = 0xe;
    else if ((digit > '@') && (digit <= 'D')) x = digit - 0x37;
    else x = digit - '0';
    current_command[0] = 0x67;
    current_command[1] = 0;
    current_command[2] = 1;
    current_command[3] = 0;
    current_command[4] = x;
    current_command[5] = 0;
    ack_or_response = 0;
    send_pkt_to_dsp ();
}

/***********************************************************/
unsigned short
TransmitLevelConverter[] = {
        0x6337, 0x586d, 0x4ecf, 0x463d,
        0x3e9a, 0x37cb, 0x31ba, 0x2c52,
        0x2780, 0x2334, 0x1f60, 0x1bf6,
        0x18ec, 0x1636, 0x13cb, 0x11a4,
        0x0fbb, 0x0000
};

char
HomologationGetTxLevel(int p1, int p2)
{
    int x; short y;
    if (p1) {
	switch (p1) {	/* These Tx levels are the defaults */
	    case 1: x =  9; break;
	    case 2: x = 10; break;
	    case 3: x = 10; break;
	    case 4: x = 10; break;
	    default: x = 10;
	}
	y = TransmitLevelConverter[x];
	if (p2 == 0)
	    return (y>>8) & 0xff;
	else
	    return (y & 0xff);
    }
    if (p2)
	x = 9;
    else
	x = 12;
    return x*10;
}


void
ModemSetTxLevel(int n)
{
    current_command[0] = 0x7b;
    current_command[1] = 0;
    current_command[2] = 4;
    current_command[3] = 0;
    if (n == 0) {
	current_command[4] = HomologationGetTxLevel(1,1);
	current_command[5] = HomologationGetTxLevel(1,0);
	ack_or_response = 0;
	send_pkt_to_dsp ();

	current_command[0] = 0x88;
	current_command[1] = 0;
	current_command[2] = 2;
	current_command[3] = 0;
	current_command[4] = HomologationGetTxLevel(0,0);
	current_command[5] = HomologationGetTxLevel(0,1);
    } else {
	current_command[4] = HomologationGetTxLevel(n,1);
	current_command[5] = HomologationGetTxLevel(n,0);
    }
    ack_or_response = 0;
    send_pkt_to_dsp ();
}

void
init_call_progress(int n)
{
    if (n < 3) {
	if ((n == 0) && call_progress_state)
	    ResetDspIdle();
	current_command[0] = 0x65;
	current_command[1] = 0;
	current_command[2] = 1;
	current_command[3] = 0;
	current_command[4] = n;
	current_command[5] = 1;
	ack_or_response = 0;
	send_pkt_to_dsp ();
    }
    call_progress_state = n;
}

void
gen_tone(int tone)
{
    ModemSetTxLevel(0);
    current_command[0] = 0x7c;
    current_command[1] = 0;
    current_command[2] = 6;
    current_command[3] = 0;
    current_command[4] = (tone >> 8) & 0xff;
    current_command[5] = tone & 0xff;
    current_command[6] = (tone >> 24) & 0xff;
    current_command[7] = (tone >> 16) & 0xff;
    current_command[8] = 1;
    current_command[9] = 0;
    ack_or_response = 0;
    send_pkt_to_dsp ();

    delay(1000);
    
    current_command[0] = 0x7c;
    current_command[1] = 0;
    current_command[2] = 6;
    current_command[3] = 0;
    current_command[4] = 0;
    current_command[5] = 0;
    current_command[6] = 0;
    current_command[7] = 0;
    current_command[8] = 0;
    current_command[9] = 0;
    ack_or_response = 0;
    send_pkt_to_dsp();

    ResetDspIdle();
    init_call_progress(0);    
}

/* ************************************************** */
void
packetize_parameter(unsigned char x, int cidx, int shift)
{
    current_command[cidx] |= x << shift;
}

void
mt_connect(int orig, int mt_bitrate, int conn_mode, int conn_clock_code,
	   int conn_leased, int conn_automode, int conn_fax,
	   int conn_calltone, int conn_etc)
{
    char reconnect = 0;
    char v23X = 0;
    char retrain = 0;	/* Retrains are not supported yet */

    if (orig) {
	if (JM_received) reconnect = 1;
    } else {
	if (CM_received) reconnect = 2;
    }
    JM_received = CM_received = 0;	/* BUG FIX */

    current_command[0] = 0x68;
    current_command[1] = 0;
    current_command[2] = 4;
    current_command[3] = 0;

    packetize_parameter(mt_bitrate, 4, 0);
    packetize_parameter(conn_clock_code, 4, 6);
    packetize_parameter(conn_mode, 5, 0);
    packetize_parameter(0x0, 5, 6);
    packetize_parameter(conn_calltone, 6, 0);
    packetize_parameter(orig, 6, 2);
    packetize_parameter(conn_automode, 6, 3);
    packetize_parameter(conn_fax, 6, 5);
    packetize_parameter(0x0, 6, 7);
    packetize_parameter(reconnect, 7, 0);
    packetize_parameter(v23X, 7, 2);
    packetize_parameter(retrain, 7, 3);
    packetize_parameter(0x0, 7, 4);
    packetize_parameter(0x0, 7, 5);
    
    ack_or_response = 0;
    send_pkt_to_dsp ();
}

void
mt_set_rate_seq(short rseq1, short rseq2, short rseq3)
{
    mt_RateSeq1 = rseq1;
    mt_RateSeq2 = rseq2;
    mt_RateSeq3 = rseq3;
    if (rseq3) {
	mt_rate_seq_map = rseq3;
	mt_set_rate_seq_param3 = 1;
    } else {
	mt_rate_seq_map = rseq1;
	mt_set_rate_seq_param3 = 0;
    }
    if (rseq2 == 0) rseq2 = rseq1;
    
    current_command[0] = 0x64;
    current_command[1] = 0;
    current_command[2] = 6;
    current_command[3] = 0;
    current_command[4] = rseq1 & 0xff;
    current_command[5] = (rseq1 >> 8) & 0xff;
    current_command[6] = rseq2 & 0xff;
    current_command[7] = (rseq2 >> 8) & 0xff;
    current_command[8] = rseq3 & 0xff;
    current_command[9] = (rseq3 >> 8) & 0xff;
    
    ack_or_response = 0;
    send_pkt_to_dsp ();
}

unsigned char
get_receiver_mode(void)
{
    if ((read_dsp_modem_status_reg() & 0xf) != 3) {
	current_command[0] = 0x72;
	current_command[1] = 0;
	current_command[2] = 0;
	current_command[3] = 0;
	ack_or_response = 1;
	send_pkt_to_dsp ();
	line_mode = dsp_response_data[0];
	ack_or_response = 0;
    }
    return line_mode;
}

void
mt_get_receiver_speed(char *tx, char *rx)
{
    char r1, r2;

    if ((line_mode == 0x11) || (line_mode == 0x12)
	|| (line_mode == 0x16) || (line_mode == 0x17)) {
	current_command[0] = 0x91;
	current_command[1] = 0;
	current_command[2] = 0;
	current_command[3] = 0;
	ack_or_response = 1;
	send_pkt_to_dsp ();
	if ((line_mode == 0x16) || (line_mode == 0x17)) {
	    r1 = dsp_response_data[6];
	    r2 = dsp_response_data[7];
	    if (r1 >= 9) r1+= 7;
	    else r1+= 3;
	    r2 += 0x1f;
	} else {
	    r1 = (dsp_response_data[0] >> 2) & 0xf;
	    r2 = (((dsp_response_data[1]<<8) |dsp_response_data[0])>>6)& 0xf;
	    if (r2 >= 9) r2+= 7;
	    else r2+= 3;
	    if (r1 >= 9) r1+= 7;
	    else r1+= 3;
	    /* if (line_mode == 0x11) r2 = r1; */
	}
    } else {
	current_command[0] = 0x70;
	current_command[1] = 0;
	current_command[2] = 0;
	current_command[3] = 0;
	ack_or_response = 1;
	send_pkt_to_dsp();
	r1 = r2 = dsp_response_data[0];
    }
    ack_or_response = 0;
    if ((mt_conn_orig==0) && ((line_mode == 0x11) || (line_mode == 0x12))) {
	*tx = r2;
	*rx = r1;
    } else {
	*tx = r1;
	*rx = r2;
    }
    line_bit_rate = *rx;
}

void
line_port_init(int n)
{
    char x;

    if (n == 0) {
	x = 0;
	current_command[0] = 0x61;
	current_command[1] = 0;
	current_command[2] = 1;
	current_command[3] = 0;
	current_command[4] = x;
	ack_or_response = 0;
	send_pkt_to_dsp();
    } else
	Msg("Not implemented yet.\n");
}

void
line_init(void)
{
    line_disable_interrupts();

    if (critical) Dbg("ERROR: Critical...\n");
    line_io_enabled = 0;
    dce_length_semaphore = 0;
    dce_status_semaphore = 0;
    
    dce_tx_interrupt_expected = 0;
    dce_rx_count = dce_tx_count = 0;
    dce_tx_remove = 0;
    dce_rx_insert = 0;
    stat_tx_count = stat_rx_count = 0;
    dte_rx_remove = 0;
    dte_tx_insert = 0;
    dce_rx_errors = 0;

    line_enable_interrupts();
}

/* ************************************************** */
void
line_io_enable(int n)
{
    if (n==1) {
	line_io_enabled = 1;
    } else
	line_io_enabled = 0;
}

int tx_room(void)
{
    return BUF_SIZE - dce_tx_count;
}

int tx_count(void)
{
    return dce_tx_count;
}

int rx_count(void)
{
    return dce_rx_count;
}

int get_rx_count(void)
{
    return stat_rx_count;
}

int get_tx_count(void)
{
    return stat_tx_count;
}

int
read_char(void)
{
    int c;
    
    line_disable_interrupts();
    if (critical) Dbg("ERROR: Critical...\n");
    
    /* if (dte_rx_remove != dce_rx_insert) { */
    if (dce_rx_count) {
	if (dte_rx_remove == dce_rx_insert) Dbg("FATAL ERROR: read_char()\n");
	c = dce_rx_buffer[dte_rx_remove++];
	if (dte_rx_remove == BUF_SIZE) dte_rx_remove = 0;
	dce_rx_count--;
    } else
	c = -1;
    
    line_enable_interrupts();
    return c;
}

int
write_char(unsigned char c)
{
    int tx_intr = 0;
    int r;
    
    line_disable_interrupts();
    if (critical) Dbg("ERROR: Critical...\n");
    
    if (dce_tx_count >= BUF_SIZE)
	r = -1;
    else {
	if ((dce_tx_count == 0) && (dce_tx_interrupt_expected == 0))
	    tx_intr = 1;
	dce_tx_buffer[dte_tx_insert++] = c;
	dce_tx_count++;
	if (dte_tx_insert == BUF_SIZE) dte_tx_insert = 0;
	
	r = c;
	if (tx_intr) {
	    dspdrv_ping_cram_intr();
	    dce_tx_interrupt_expected = 1;
	}
    }

    line_enable_interrupts();
    return r;
}

void
print_io_stats(int detail_level)
{
    Msg("STATISTICS:\n");
    Msg("RX chars  : %d\n", get_rx_count());
    Msg("TX chars  : %d\n", get_tx_count());
    Msg("RX errors : %d ?\n", dce_rx_errors);

#ifdef DEBUG
    Dbg("TX COUNT = %d\n", dce_tx_count);
    Dbg("RX COUNT = %d\n", dce_rx_count);
    
    if (detail_level >= 1) {
	int i;
	Dbg("IO STATS:\n");
	Dbg("             RX       TX\n");
	for(i=0; i<32; i++)
	    Dbg(" (%2d) --> %5d    %5d\n", i+1, RX_stat[i], TX_stat[i]);
    }
#endif
}

#define WAIT_FOR_V8_TIME   5
    
int
v8_orig(unsigned char bitrate)
{
    char cp, last_cp, xs, xl;
    int i, t;
    
    sending_CM = sending_JM = 0;
    JM_received = CM_received = 0;
    CM_count = JM_count = 0;
    v8_tx_msg_size = 6;
    
    last_cp = call_progress();
    Dbg("V8 Orig\n");
    
    for (t=0; t<WAIT_FOR_V8_TIME*100; t++) {
	if ((read_dsp_modem_status_reg() & 0xf) == 2)
	    return 1;	/* connected */
	cp = call_progress();
	if (cp != last_cp) {
	    last_cp = cp;
	    if (cp == 0x22) {
		Dbg("ANSam tone detected.\n");	/* ? XX */
	    } else if ((cp == 0x1f) || (cp == 0x20)) {
		/* Dbg("Te expired?\n"); */
		v8_configured = 1;
		Dbg("Sending CM...\n");
		sending_CM = 1;
		v8_tx_buffer[0] = 0xff;	/* ONEs */
		v8_tx_buffer[1] = 0xe0;	/* CM sync */
		v8_tx_buffer[2] = 0xc1;	/* call function */
		/* modulation modes */
		v8_tx_buffer[3] = 0x45;	/* V.34 duplex  */
		v8_tx_buffer[4] = 0x13;	/* V.32, V.22   */
		v8_tx_buffer[5] = 0x94;	/* V.21, V.23   */
	    
		dspdrv_WriteCram(0x54, v8_tx_buffer, v8_tx_msg_size);
		xs = 0;
		dspdrv_WriteCram(0x75, &xs, 1);
		dspdrv_WriteCram(0x74, &v8_tx_msg_size, 1);
		jm_seq_match = 0;
		v8_status = 0;
		line_io_enabled = 1;
	    
		while (!jm_seq_match) {
		    if (CM_count > 20) {
			Dbg("V8 CM count limit exceeded\n");
			sending_CM = 0;
			v8_configured = 0;
			line_io_enabled = 0;
			return -1;
		    }
		    delay (10); 	/* 10ms */
		}

		Dbg("CM count: %d\n", CM_count);
		JM_received = 1;
		sending_CM = 0;
		Dbg("JM received: ( ");
		for (i=0; i<v8_rx_msg_size; i++)
		    Dbg("%x ", dce_rx_buffer[i]);
		Dbg(")\n");
		
		v8_tx_buffer[0] = 0;
		v8_tx_buffer[1] = 0;
		v8_tx_buffer[2] = 0;
		Dbg("Sending CJ...\n");
		xl = 3;
		dspdrv_WriteCram(0x54, v8_tx_buffer, 3);
		xs = 0;
		dspdrv_WriteCram(0x75, &xs, 1);
		dspdrv_WriteCram(0x74, &xl, 1);
	    
		/* process_JM_CM() */
		
		Dbg("Reconnecting...\n");
		mt_connect(M_CONN_ORIG, bitrate, MD_V34, 0, 0, 3, 0, 00, 0);
		v8_configured = 0;
		/* line_io_enabled = 0; */
		return 1;
	    } else
		Msg("New Call Progress : %x\n", cp);
	}
	delay(10);	/* 10 ms */
    }
    return -1;
}
